@@ -264,6 +264,28 @@ def metrics_by_tag(tag_name):
     return metrics_by_name_list(names)
 
 
+def untag(name, tag_name):
+    """
+    Remove the given tag from the given metric.
+    Return True if the metric was tagged, False otherwise
+    """
+
+    with LOCK:
+        by_tag = TAGS.get(tag_name, None)
+        if not by_tag:
+            return False
+        try:
+            by_tag.remove(name)
+
+            # remove the tag if no associations left
+            if not by_tag:
+                TAGS.pop(tag_name)
+
+            return True
+        except KeyError:
+            return False
+
+
 def metrics_by_name_list(names):
     """
     Return a dictionary with {metric name: metric value} for all the metrics with the given names.
